@@ -2,6 +2,7 @@
  *  ARMv4 code generator for TCC
  * 
  *  Copyright (c) 2003 Daniel Glckner
+ *  Copyright (c) 2012 Thomas Preud'homme
  *
  *  Based on i386-gen.c by Fabrice Bellard
  *
@@ -20,10 +21,13 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#ifdef TARGET_DEFS_ONLY
+
 #ifdef TCC_ARM_EABI
+#ifndef TCC_ARM_VFP // Avoid useless warning
 #define TCC_ARM_VFP
 #endif
-
+#endif
 
 /* number of available registers */
 #ifdef TCC_ARM_VFP
@@ -32,6 +36,10 @@
 #define NB_REGS             9
 #endif
 
+#ifndef TCC_ARM_VERSION
+# define TCC_ARM_VERSION 5
+#endif
+
 /* a register can belong to several classes. The classes must be
    sorted from more general to more precise (see gv2() code which does
    assumptions on it). */
@@ -75,32 +83,6 @@ enum {
 #endif
 };
 
-int reg_classes[NB_REGS] = {
-    /* r0 */ RC_INT | RC_R0,
-    /* r1 */ RC_INT | RC_R1,
-    /* r2 */ RC_INT | RC_R2,
-    /* r3 */ RC_INT | RC_R3,
-    /* r12 */ RC_INT | RC_R12,
-    /* f0 */ RC_FLOAT | RC_F0,
-    /* f1 */ RC_FLOAT | RC_F1,
-    /* f2 */ RC_FLOAT | RC_F2,
-    /* f3 */ RC_FLOAT | RC_F3,
-#ifdef TCC_ARM_VFP
- /* d4/s8 */ RC_FLOAT | RC_F4,
-/* d5/s10 */ RC_FLOAT | RC_F5,
-/* d6/s12 */ RC_FLOAT | RC_F6,
-/* d7/s14 */ RC_FLOAT | RC_F7,
-#endif
-};
-
-static int two2mask(int a,int b) {
-  return (reg_classes[a]|reg_classes[b])&~(RC_INT|RC_FLOAT);
-}
-
-static int regmask(int r) {
-  return reg_classes[r]&~(RC_INT|RC_FLOAT);
-}
-
 #ifdef TCC_ARM_VFP
 #define T2CPR(t) (((t) & VT_BTYPE) != VT_FLOAT ? 0x100 : 0)
 #endif
@@ -124,15 +106,6 @@ static int regmask(int r) {
    are directly pushed on stack. */
 //#define FUNC_STRUCT_PARAM_AS_PTR
 
-#if defined(TCC_ARM_EABI) && defined(TCC_ARM_VFP)
-static CType float_type, double_type, func_float_type, func_double_type;
-#define func_ldouble_type func_double_type
-#else
-#define func_float_type func_old_type
-#define func_double_type func_old_type
-#define func_ldouble_type func_old_type
-#endif
-
 /* pointer size, in bytes */
 #define PTR_SIZE 4
 
@@ -163,6 +136,7 @@ static CType float_type, double_type, func_float_type, func_double_type;
 
 /* relocation type for 32 bit data relocation */
 #define R_DATA_32   R_ARM_ABS32
+#define R_DATA_PTR  R_ARM_ABS32
 #define R_JMP_SLOT  R_ARM_JUMP_SLOT
 #define R_COPY      R_ARM_COPY
 
@@ -170,17 +144,67 @@ static CType float_type, double_type, func_float_type, func_double_type;
 #define ELF_PAGE_SIZE  0x1000
 
 /******************************************************/
-static unsigned long func_sub_sp_offset,last_itod_magic;
+#else /* ! TARGET_DEFS_ONLY */
+/******************************************************/
+#include "tcc.h"
+
+ST_DATA const int reg_classes[NB_REGS] = {
+    /* r0 */ RC_INT | RC_R0,
+    /* r1 */ RC_INT | RC_R1,
+    /* r2 */ RC_INT | RC_R2,
+    /* r3 */ RC_INT | RC_R3,
+    /* r12 */ RC_INT | RC_R12,
+    /* f0 */ RC_FLOAT | RC_F0,
+    /* f1 */ RC_FLOAT | RC_F1,
+    /* f2 */ RC_FLOAT | RC_F2,
+    /* f3 */ RC_FLOAT | RC_F3,
+#ifdef TCC_ARM_VFP
+ /* d4/s8 */ RC_FLOAT | RC_F4,
+/* d5/s10 */ RC_FLOAT | RC_F5,
+/* d6/s12 */ RC_FLOAT | RC_F6,
+/* d7/s14 */ RC_FLOAT | RC_F7,
+#endif
+};
+
+static int func_sub_sp_offset, last_itod_magic;
 static int leaffunc;
 
-void o(unsigned long i)
+#if defined(TCC_ARM_EABI) && defined(TCC_ARM_VFP)
+static CType float_type, double_type, func_float_type, func_double_type;
+ST_FUNC void arm_init_types(void)
+{
+    float_type.t = VT_FLOAT;
+    double_type.t = VT_DOUBLE;
+    func_float_type.t = VT_FUNC;
+    func_float_type.ref = sym_push(SYM_FIELD, &float_type, FUNC_CDECL, FUNC_OLD);
+    func_double_type.t = VT_FUNC;
+    func_double_type.ref = sym_push(SYM_FIELD, &double_type, FUNC_CDECL, FUNC_OLD);
+}
+#else
+#define func_float_type func_old_type
+#define func_double_type func_old_type
+#define func_ldouble_type func_old_type
+ST_FUNC void arm_init_types(void) {}
+#endif
+
+static int two2mask(int a,int b) {
+  return (reg_classes[a]|reg_classes[b])&~(RC_INT|RC_FLOAT);
+}
+
+static int regmask(int r) {
+  return reg_classes[r]&~(RC_INT|RC_FLOAT);
+}
+
+/******************************************************/
+
+void o(uint32_t i)
 {
   /* this is a good place to start adding big-endian support*/
   int ind1;
 
   ind1 = ind + 4;
   if (!cur_text_section)
-    error("compiler error! This happens f.ex. if the compiler\n"
+    tcc_error("compiler error! This happens f.ex. if the compiler\n"
          "can't evaluate constant expressions outside of a function.");
   if (ind1 > cur_text_section->data_allocated)
     section_realloc(cur_text_section, ind1);
@@ -193,10 +217,10 @@ void o(unsigned long i)
   cur_text_section->data[ind++] = i;
 }
 
-static unsigned long stuff_const(unsigned long op,unsigned long c)
+static uint32_t stuff_const(uint32_t op, uint32_t c)
 {
   int try_neg=0;
-  unsigned long nc = 0,negop = 0;
+  uint32_t nc = 0, negop = 0;
 
   switch(op&0x1F00000)
   {
@@ -230,7 +254,7 @@ static unsigned long stuff_const(unsigned long op,unsigned long c)
       break;
   }
   do {
-    unsigned long m;
+    uint32_t m;
     int i;
     if(c<256) /* catch undefined <<32 */
       return op|c;
@@ -247,13 +271,13 @@ static unsigned long stuff_const(unsigned long op,unsigned long c)
 
 
 //only add,sub
-void stuff_const_harder(unsigned long op,unsigned long v) {
-  unsigned long x;
+void stuff_const_harder(uint32_t op, uint32_t v) {
+  uint32_t x;
   x=stuff_const(op,v);
   if(x)
     o(x);
   else {
-    unsigned long a[16],nv,no,o2,n2;
+    uint32_t a[16], nv, no, o2, n2;
     int i,j,k;
     a[0]=0xff;
     o2=(op&0xfff0ffff)|((op&0xf000)<<4);;
@@ -303,13 +327,13 @@ void stuff_const_harder(unsigned long op,unsigned long v) {
   }
 }
 
-unsigned long encbranch(int pos,int addr,int fail)
+ST_FUNC uint32_t encbranch(int pos, int addr, int fail)
 {
   addr-=pos+8;
   addr/=4;
   if(addr>=0x1000000 || addr<-0x1000000) {
     if(fail)
-      error("FIXME: function bigger than 32MB");
+      tcc_error("FIXME: function bigger than 32MB");
     return 0;
   }
   return 0x0A000000|(addr&0xffffff);
@@ -318,7 +342,7 @@ unsigned long encbranch(int pos,int addr,int fail)
 int decbranch(int pos)
 {
   int x;
-  x=*(int *)(cur_text_section->data + pos);
+  x=*(uint32_t *)(cur_text_section->data + pos);
   x&=0x00ffffff;
   if(x&0x800000)
     x-=0x1000000;
@@ -328,10 +352,10 @@ int decbranch(int pos)
 /* output a symbol and patch all calls to it */
 void gsym_addr(int t, int a)
 {
-  unsigned long *x;
+  uint32_t *x;
   int lt;
   while(t) {
-    x=(unsigned long *)(cur_text_section->data + t);
+    x=(uint32_t *)(cur_text_section->data + t);
     t=decbranch(lt=t);
     if(a==lt+4)
       *x=0xE1A00000; // nop
@@ -348,34 +372,34 @@ void gsym(int t)
 }
 
 #ifdef TCC_ARM_VFP
-static unsigned long vfpr(int r)
+static uint32_t vfpr(int r)
 {
   if(r<TREG_F0 || r>TREG_F7)
-    error("compiler error! register %i is no vfp register",r);
+    tcc_error("compiler error! register %i is no vfp register",r);
   return r-5;
 }
 #else
-static unsigned long fpr(int r)
+static uint32_t fpr(int r)
 {
   if(r<TREG_F0 || r>TREG_F3)
-    error("compiler error! register %i is no fpa register",r);
+    tcc_error("compiler error! register %i is no fpa register",r);
   return r-5;
 }
 #endif
 
-static unsigned long intr(int r)
+static uint32_t intr(int r)
 {
   if(r==4)
     return 12;
   if((r<0 || r>4) && r!=14)
-    error("compiler error! register %i is no int register",r);
+    tcc_error("compiler error! register %i is no int register",r);
   return r;
 }
 
-static void calcaddr(unsigned long *base,int *off,int *sgn,int maxoff,unsigned shift)
+static void calcaddr(uint32_t *base, int *off, int *sgn, int maxoff, unsigned shift)
 {
   if(*off>maxoff || *off&((1<<shift)-1)) {
-    unsigned long x,y;
+    uint32_t x, y;
     x=0xE280E000;
     if(*sgn)
       x=0xE240E000;
@@ -399,7 +423,7 @@ static void calcaddr(unsigned long *base,int *off,int *sgn,int maxoff,unsigned s
   }
 }
 
-static unsigned long mapcc(int cc)
+static uint32_t mapcc(int cc)
 {
   switch(cc)
   {
@@ -428,7 +452,7 @@ static unsigned long mapcc(int cc)
     case TOK_GT:
       return 0xC0000000; /* GT */
   }
-  error("unexpected condition code");
+  tcc_error("unexpected condition code");
   return 0xE0000000; /* AL */
 }
 
@@ -461,7 +485,7 @@ static int negcc(int cc)
     case TOK_GT:
       return TOK_LE;
   }
-  error("unexpected condition code");
+  tcc_error("unexpected condition code");
   return TOK_NE;
 }
 
@@ -469,7 +493,7 @@ static int negcc(int cc)
 void load(int r, SValue *sv)
 {
   int v, ft, fc, fr, sign;
-  unsigned long op;
+  uint32_t op;
   SValue v1;
 
   fr = sv->r;
@@ -485,7 +509,7 @@ void load(int r, SValue *sv)
   
   v = fr & VT_VALMASK;
   if (fr & VT_LVAL) {
-    unsigned long base=0xB; // fp
+    uint32_t base = 0xB; // fp
     if(v == VT_LLOCAL) {
       v1.type.t = VT_PTR;
       v1.r = VT_LOCAL | VT_LVAL;
@@ -601,7 +625,7 @@ void load(int r, SValue *sv)
       return;
     }
   }
-  error("load unimplemented!");
+  tcc_error("load unimplemented!");
 }
 
 /* store register 'r' in lvalue 'v' */
@@ -609,7 +633,7 @@ void store(int r, SValue *sv)
 {
   SValue v1;
   int v, ft, fc, fr, sign;
-  unsigned long op;
+  uint32_t op;
 
   fr = sv->r;
   ft = sv->type.t;
@@ -624,7 +648,7 @@ void store(int r, SValue *sv)
   
   v = fr & VT_VALMASK; 
   if (fr & VT_LVAL || fr == VT_LOCAL) {
-    unsigned long base=0xb;
+    uint32_t base = 0xb;
     if(v < VT_CONST) {
       base=intr(v);
       v=VT_LOCAL;
@@ -682,7 +706,7 @@ void store(int r, SValue *sv)
       return;
     }
   }
-  error("store unimplemented");
+  tcc_error("store unimplemented");
 }
 
 static void gadd_sp(int val)
@@ -695,7 +719,7 @@ static void gcall_or_jmp(int is_jmp)
 {
   int r;
   if ((vtop->r & (VT_VALMASK | VT_LVAL)) == VT_CONST) {
-    unsigned long x;
+    uint32_t x;
     /* constant case */
     x=encbranch(ind,ind+vtop->c.ul,0);
     if(x) {
@@ -722,22 +746,91 @@ static void gcall_or_jmp(int is_jmp)
   }
 }
 
+#ifdef TCC_ARM_HARDFLOAT
+static int is_float_hgen_aggr(CType *type)
+{
+  if ((type->t & VT_BTYPE) == VT_STRUCT) {
+    struct Sym *ref;
+    int btype, nb_fields = 0;
+
+    ref = type->ref;
+    btype = ref->type.t & VT_BTYPE;
+    if (btype == VT_FLOAT || btype == VT_DOUBLE) {
+      for(; ref && btype == (ref->type.t & VT_BTYPE); ref = ref->next, nb_fields++);
+      return !ref && nb_fields <= 4;
+    }
+  }
+  return 0;
+}
+
+struct avail_regs {
+  /* worst case: f(float, double, 3 float struct, double, 3 float struct, double) */
+  signed char avail[3];
+  int first_hole;
+  int last_hole;
+  int first_free_reg;
+};
+
+#define AVAIL_REGS_INITIALIZER (struct avail_regs) { { 0, 0, 0}, 0, 0, 0 }
+
+/* Assign a register for a CPRC param with correct size and alignment
+ * size and align are in bytes, as returned by type_size */
+int assign_fpreg(struct avail_regs *avregs, int align, int size)
+{
+  int first_reg = 0;
+
+  if (avregs->first_free_reg == -1)
+    return -1;
+  if (align >> 3) { // alignment needed (base type: double)
+    first_reg = avregs->first_free_reg;
+    if (first_reg & 1)
+      avregs->avail[avregs->last_hole++] = first_reg++;
+  } else {
+    if (size == 4 && avregs->first_hole != avregs->last_hole)
+      return avregs->avail[avregs->first_hole++];
+    else
+      first_reg = avregs->first_free_reg;
+  }
+  if (first_reg + size / 4 <= 16) {
+    avregs->first_free_reg = first_reg + size / 4;
+    return first_reg;
+  }
+  avregs->first_free_reg = -1;
+  return -1;
+}
+#endif
+
 /* Generate function call. The function address is pushed first, then
    all the parameters in call order. This functions pops all the
    parameters and the function address. */
 void gfunc_call(int nb_args)
 {
-  int size, align, r, args_size, i;
-  Sym *func_sym;
+  int size, align, r, args_size, i, ncrn, ncprn, argno, vfp_argno;
   signed char plan[4][2]={{-1,-1},{-1,-1},{-1,-1},{-1,-1}};
-  int todo=0xf, keep, plan2[4]={0,0,0,0};
+  SValue *before_stack = NULL; /* SValue before first on stack argument */
+  SValue *before_vfpreg_hfa = NULL; /* SValue before first in VFP reg hfa argument */
+#ifdef TCC_ARM_HARDFLOAT
+  struct avail_regs avregs = AVAIL_REGS_INITIALIZER;
+  signed char vfp_plan[16];
+  int plan2[4+16];
+  int variadic;
+#else
+  int plan2[4]={0,0,0,0};
+#endif
+  int vfp_todo=0;
+  int todo=0, keep;
 
+#ifdef TCC_ARM_HARDFLOAT
+  memset(vfp_plan, -1, sizeof(vfp_plan));
+  memset(plan2, 0, sizeof(plan2));
+  variadic = (vtop[-nb_args].type.ref->c == FUNC_ELLIPSIS);
+#endif
   r = vtop->r & VT_VALMASK;
   if (r == VT_CMP || (r & ~1) == VT_JMP)
     gv(RC_INT);
 #ifdef TCC_ARM_EABI
   if((vtop[-nb_args].type.ref->type.t & VT_BTYPE) == VT_STRUCT
-     && type_size(&vtop[-nb_args].type, &align) <= 4) {
+     && type_size(&vtop[-nb_args].type.ref->type, &align) <= 4) {
     SValue tmp;
     tmp=vtop[-nb_args];
     vtop[-nb_args]=vtop[-nb_args+1];
@@ -745,45 +838,137 @@ void gfunc_call(int nb_args)
     --nb_args;
   }
   
-  vpushi(0);
+  vpushi(0), nb_args++;
   vtop->type.t = VT_LLONG;
-  args_size = 0;
-  for(i = nb_args + 1 ; i-- ;) {
-    size = type_size(&vtop[-i].type, &align);
-    if(args_size & (align-1)) {
-      vpushi(0);
-      vtop->type.t = VT_VOID; /* padding */
-      vrott(i+2);
-      args_size += 4;
-      ++nb_args;
-    }
-    args_size += (size + 3) & -4;
-  }
-  vtop--;
 #endif
-  args_size = 0;
-  for(i = nb_args ; i-- && args_size < 16 ;) {
+  ncrn = ncprn = argno = vfp_argno = args_size = 0;
+  /* Assign argument to registers and stack with alignment.
+     If, considering alignment constraints, enough registers of the correct type
+     (core or VFP) are free for the current argument, assign them to it, else
+     allocate on stack with correct alignment. Whenever a structure is allocated
+     in registers or on stack, it is always put on the stack at this stage. The
+     stack is divided in 3 zones. The zone are, from low addresses to high
+     addresses: structures to be loaded in core registers, structures to be
+     loaded in VFP registers, argument allocated to stack. SValue's representing
+     structures in the first zone are moved just after the SValue pointed by
+     before_vfpreg_hfa. SValue's representing structures in the second zone are
+     moved just after the SValue pointer by before_stack. */
+  for(i = nb_args; i-- ;) {
+    int j, assigned_vfpreg = 0;
+    size = type_size(&vtop[-i].type, &align);
     switch(vtop[-i].type.t & VT_BTYPE) {
       case VT_STRUCT:
       case VT_FLOAT:
       case VT_DOUBLE:
       case VT_LDOUBLE:
-      size = type_size(&vtop[-i].type, &align);
-        size = (size + 3) & -4;
-      args_size += size;
-        break;
-      default:
-      plan[nb_args-1-i][0]=args_size/4;
-      args_size += 4;
-      if ((vtop[-i].type.t & VT_BTYPE) == VT_LLONG && args_size < 16) {
-	plan[nb_args-1-i][1]=args_size/4;
-	args_size += 4;
+#ifdef TCC_ARM_HARDFLOAT
+      if (!variadic) {
+        int hfa = 0; /* Homogeneous float aggregate */
+
+        if (is_float(vtop[-i].type.t)
+            || (hfa = is_float_hgen_aggr(&vtop[-i].type))) {
+          int end_reg;
+
+          assigned_vfpreg = assign_fpreg(&avregs, align, size);
+          end_reg = assigned_vfpreg + (size - 1) / 4;
+          if (assigned_vfpreg >= 0) {
+            vfp_plan[vfp_argno++]=TREG_F0 + assigned_vfpreg/2;
+            if (hfa) {
+              /* before_stack can only have been set because all core registers
+                 are assigned, so no need to care about before_vfpreg_hfa if
+                 before_stack is set */
+              if (before_stack) {
+	        vrote(&vtop[-i], &vtop[-i] - before_stack);
+                before_stack++;
+              } else if (!before_vfpreg_hfa)
+                before_vfpreg_hfa = &vtop[-i-1];
+              for (j = assigned_vfpreg; j <= end_reg; j++)
+                vfp_todo|=(1<<j);
+            }
+            continue;
+          } else {
+            if (!hfa)
+              vfp_argno++;
+            /* No need to update before_stack as no more hfa can be allocated in
+               VFP regs */
+            if (!before_vfpreg_hfa)
+              before_vfpreg_hfa = &vtop[-i-1];
+            break;
+          }
+        }
       }
+#endif
+      ncrn = (ncrn + (align-1)/4) & -(align/4);
+      size = (size + 3) & -4;
+      if (ncrn + size/4 <= 4 || (ncrn < 4 && assigned_vfpreg != -1)) {
+        /* Either there is HFA in VFP registers, or there is arguments on stack,
+           it cannot be both. Hence either before_stack already points after
+           the slot where the vtop[-i] SValue is moved, or before_stack will not
+           be used */
+        if (before_vfpreg_hfa) {
+	  vrote(&vtop[-i], &vtop[-i] - before_vfpreg_hfa);
+          before_vfpreg_hfa++;
+        }
+        for (j = ncrn; j < 4 && j < ncrn + size / 4; j++)
+          todo|=(1<<j);
+        ncrn+=size/4;
+        if (ncrn > 4) {
+          args_size = (ncrn - 4) * 4;
+          if (!before_stack)
+            before_stack = &vtop[-i-1];
+        }
+      }
+      else {
+        ncrn = 4;
+        /* No need to set before_vfpreg_hfa if not set since there will no
+           longer be any structure assigned to core registers */
+        if (!before_stack)
+          before_stack = &vtop[-i-1];
+        break;
+      }
+      continue;
+      default:
+#ifdef TCC_ARM_EABI
+      if (!i) {
+        break;
+      }
+#endif
+      if (ncrn < 4) {
+        int is_long = (vtop[-i].type.t & VT_BTYPE) == VT_LLONG;
+
+        if (is_long) {
+          ncrn = (ncrn + 1) & -2;
+          if (ncrn == 4) {
+            argno++;
+            break;
+          }
+        }
+        plan[argno++][0]=ncrn++;
+        if (is_long) {
+          plan[argno-1][1]=ncrn++;
+        }
+        continue;
+      }
+      argno++;
     }
+#ifdef TCC_ARM_EABI
+    if(args_size & (align-1)) {
+      vpushi(0);
+      vtop->type.t = VT_VOID; /* padding */
+      vrott(i+2);
+      args_size += 4;
+      nb_args++;
+      argno++;
+    }
+#endif
+    args_size += (size + 3) & -4;
   }
+#ifdef TCC_ARM_EABI
+  vtop--, nb_args--;
+#endif
   args_size = keep = 0;
   for(i = 0;i < nb_args; i++) {
-    vnrott(keep+1);
+    vrotb(keep+1);
     if ((vtop->type.t & VT_BTYPE) == VT_STRUCT) {
       size = type_size(&vtop->type, &align);
       /* align to stack align size */
@@ -799,6 +984,12 @@ void gfunc_call(int nb_args)
       vtop--;
       args_size += size;
     } else if (is_float(vtop->type.t)) {
+#ifdef TCC_ARM_HARDFLOAT
+      if (!variadic && --vfp_argno<16 && vfp_plan[vfp_argno]!=-1) {
+        plan2[keep++]=vfp_plan[vfp_argno];
+        continue;
+      }
+#endif
 #ifdef TCC_ARM_VFP
       r=vfpr(gv(RC_FLOAT))<<12;
       size=4;
@@ -833,57 +1024,61 @@ void gfunc_call(int nb_args)
       size=4;
       if ((vtop->type.t & VT_BTYPE) == VT_LLONG) {
 	lexpand_nr();
-	s=RC_INT;
-	if(nb_args-i<5 && plan[nb_args-i-1][1]!=-1) {
-	  s=regmask(plan[nb_args-i-1][1]);
-	  todo&=~(1<<plan[nb_args-i-1][1]);
-	}
-	if(s==RC_INT) {
-	  r = gv(s);
+	s=-1;
+	if(--argno<4 && plan[argno][1]!=-1)
+	  s=plan[argno][1];
+	argno++;
+	size = 8;
+	if(s==-1) {
+	  r = gv(RC_INT);
 	  o(0xE52D0004|(intr(r)<<12)); /* str r,[sp,#-4]! */
 	  vtop--;
 	} else {
+	  size=0;
 	  plan2[keep]=s;
 	  keep++;
           vswap();
 	}
-	size = 8;
-      }
-      s=RC_INT;
-      if(nb_args-i<5 && plan[nb_args-i-1][0]!=-1) {
-        s=regmask(plan[nb_args-i-1][0]);
-	todo&=~(1<<plan[nb_args-i-1][0]);
       }
+      s=-1;
+      if(--argno<4 && plan[argno][0]!=-1)
+        s=plan[argno][0];
 #ifdef TCC_ARM_EABI
       if(vtop->type.t == VT_VOID) {
-        if(s == RC_INT)
+        if(s == -1)
           o(0xE24DD004); /* sub sp,sp,#4 */
         vtop--;
       } else
-#endif      
-      if(s == RC_INT) {
-	r = gv(s);
+#endif
+      if(s == -1) {
+	r = gv(RC_INT);
 	o(0xE52D0004|(intr(r)<<12)); /* str r,[sp,#-4]! */
 	vtop--;
       } else {
+        size=0;
 	plan2[keep]=s;
 	keep++;
       }
       args_size += size;
     }
   }
-  for(i=keep;i--;) {
-    gv(plan2[i]);
-    vrott(keep);
+  for(i = 0; i < keep; i++) {
+    vrotb(keep);
+    gv(regmask(plan2[i]));
+#ifdef TCC_ARM_HARDFLOAT
+    /* arg is in s(2d+1): plan2[i]<plan2[i+1] => alignment occured (ex f,d,f) */
+    if (i < keep - 1 && is_float(vtop->type.t) && (plan2[i] <= plan2[i + 1])) {
+      o(0xEEF00A40|(vfpr(plan2[i])<<12)|vfpr(plan2[i]));
+    }
+#endif
   }
 save_regs(keep); /* save used temporary registers */
   keep++;
-  if(args_size) {
-    int n;
-    n=args_size/4;
-    if(n>4)
-      n=4;
-    todo&=((1<<n)-1);
+  if(ncrn) {
+    int nb_regs=0;
+    if (ncrn>4)
+      ncrn=4;
+    todo&=((1<<ncrn)-1);
     if(todo) {
       int i;
       o(0xE8BD0000|todo);
@@ -892,12 +1087,31 @@ save_regs(keep); /* save used temporary registers */
 	  vpushi(0);
 	  vtop->r=i;
 	  keep++;
+	  nb_regs++;
 	}
     }
-    args_size-=n*4;
+    args_size-=nb_regs*4;
   }
-  vnrott(keep);
-  func_sym = vtop->type.ref;
+  if(vfp_todo) {
+    int nb_fregs=0;
+
+    for(i=0;i<16;i++)
+      if(vfp_todo&(1<<i)) {
+        o(0xED9D0A00|(i&1)<<22|(i>>1)<<12|nb_fregs);
+        vpushi(0);
+        /* There might be 2 floats in a double VFP reg but that doesn't seem
+           to matter */
+        if (!(i%2))
+          vtop->r=TREG_F0+i/2;
+        keep++;
+        nb_fregs++;
+      }
+    if (nb_fregs) {
+      gadd_sp(nb_fregs*4);
+      args_size-=nb_fregs*4;
+    }
+  }
+  vrotb(keep);
   gcall_or_jmp(0);
   if (args_size)
       gadd_sp(args_size);
@@ -909,7 +1123,11 @@ save_regs(keep); /* save used temporary registers */
     ++keep;
   }
 #ifdef TCC_ARM_VFP
+#ifdef TCC_ARM_HARDFLOAT
+  else if(variadic && is_float(vtop->type.ref->type.t)) {
+#else
   else if(is_float(vtop->type.ref->type.t)) {
+#endif
     if((vtop->type.ref->type.t & VT_BTYPE) == VT_FLOAT) {
       o(0xEE000A10); /* fmsr s0,r0 */
     } else {
@@ -927,26 +1145,37 @@ save_regs(keep); /* save used temporary registers */
 void gfunc_prolog(CType *func_type)
 {
   Sym *sym,*sym2;
-  int n,addr,size,align;
+  int n,nf,size,align, variadic, struct_ret = 0;
+#ifdef TCC_ARM_HARDFLOAT
+  struct avail_regs avregs = AVAIL_REGS_INITIALIZER;
+#endif
 
   sym = func_type->ref;
   func_vt = sym->type;
-  
-  n = 0;
-  addr = 0;
+
+  n = nf = 0;
+  variadic = (func_type->ref->c == FUNC_ELLIPSIS);
   if((func_vt.t & VT_BTYPE) == VT_STRUCT
      && type_size(&func_vt,&align) > 4)
   {
-    func_vc = addr;
-    addr += 4;
     n++;
+    struct_ret = 1;
+    func_vc = 12; /* Offset from fp of the place to store the result */
   }
-  for(sym2=sym->next;sym2 && n<4;sym2=sym2->next) {
+  for(sym2=sym->next;sym2 && (n<4 || nf<16);sym2=sym2->next) {
     size = type_size(&sym2->type, &align);
-    n += (size + 3) / 4;
+#ifdef TCC_ARM_HARDFLOAT
+    if (!variadic && (is_float(sym2->type.t)
+        || is_float_hgen_aggr(&sym2->type))) {
+      int tmpnf = assign_fpreg(&avregs, align, size) + 1;
+      nf = (tmpnf > nf) ? tmpnf : nf;
+    } else
+#endif
+    if (n < 4)
+      n += (size + 3) / 4;
   }
   o(0xE1A0C00D); /* mov ip,sp */
-  if(func_type->ref->c == FUNC_ELLIPSIS)
+  if(variadic)
     n=4;
   if(n) {
     if(n>4)
@@ -956,32 +1185,72 @@ void gfunc_prolog(CType *func_type)
 #endif
     o(0xE92D0000|((1<<n)-1)); /* save r0-r4 on stack if needed */
   }
+  if (nf) {
+    if (nf>16)
+      nf=16;
+    nf=(nf+1)&-2; /* nf => HARDFLOAT => EABI */
+    o(0xED2D0A00|nf); /* save s0-s15 on stack if needed */
+  }
   o(0xE92D5800); /* save fp, ip, lr */
-  o(0xE28DB00C); /* add fp, sp, #12 */
+  o(0xE1A0B00D); /* mov fp, sp */
   func_sub_sp_offset = ind;
-  o(0xE1A00000); /* nop, leave space for stack adjustment */
-  while ((sym = sym->next)) {
-    CType *type;
-    type = &sym->type;
-    size = type_size(type, &align);
-    size = (size + 3) & -4;
-#ifdef TCC_ARM_EABI
-    addr = (addr + align - 1) & -align;
+  o(0xE1A00000); /* nop, leave space for stack adjustment in epilogue */
+  {
+    int addr, pn = struct_ret, sn = 0; /* pn=core, sn=stack */
+
+#ifdef TCC_ARM_HARDFLOAT
+    avregs = AVAIL_REGS_INITIALIZER;
 #endif
-    sym_push(sym->v & ~SYM_FIELD, type, VT_LOCAL | lvalue_type(type->t), addr);
-    addr += size;
+    while ((sym = sym->next)) {
+      CType *type;
+      type = &sym->type;
+      size = type_size(type, &align);
+      size = (size + 3) >> 2;
+      align = (align + 3) & ~3;
+#ifdef TCC_ARM_HARDFLOAT
+      if (!variadic && (is_float(sym->type.t)
+          || is_float_hgen_aggr(&sym->type))) {
+        int fpn = assign_fpreg(&avregs, align, size << 2);
+        if (fpn >= 0) {
+          addr = fpn * 4;
+        } else
+          goto from_stack;
+      } else
+#endif
+      if (pn < 4) {
+#ifdef TCC_ARM_EABI
+        pn = (pn + (align-1)/4) & -(align/4);
+#endif
+        addr = (nf + pn) * 4;
+        pn += size;
+        if (!sn && pn > 4)
+          sn = (pn - 4);
+      } else {
+#ifdef TCC_ARM_HARDFLOAT
+from_stack:
+#endif
+#ifdef TCC_ARM_EABI
+        sn = (sn + (align-1)/4) & -(align/4);
+#endif
+        addr = (n + nf + sn) * 4;
+        sn += size;
+      }
+      sym_push(sym->v & ~SYM_FIELD, type, VT_LOCAL | lvalue_type(type->t), addr+12);
+    }
   }
   last_itod_magic=0;
   leaffunc = 1;
-  loc = -12;
+  loc = 0;
 }
 
 /* generate function epilog */
 void gfunc_epilog(void)
 {
-  unsigned long x;
+  uint32_t x;
   int diff;
 #ifdef TCC_ARM_EABI
+  /* Useless but harmless copy of the float result into main register(s) in case
+     of variadic function in the hardfloat variant */
   if(is_float(func_vt.t)) {
     if((func_vt.t & VT_BTYPE) == VT_FLOAT)
       o(0xEE100A10); /* fmrs r0, s0 */
@@ -991,24 +1260,24 @@ void gfunc_epilog(void)
     }
   }
 #endif
-  o(0xE91BA800); /* restore fp, sp, pc */
+  o(0xE89BA800); /* restore fp, sp, pc */
   diff = (-loc + 3) & -4;
 #ifdef TCC_ARM_EABI
   if(!leaffunc)
-    diff = (diff + 7) & -8;
+    diff = ((diff + 11) & -8) - 4;
 #endif
-  if(diff > 12) {
+  if(diff > 0) {
     x=stuff_const(0xE24BD000, diff); /* sub sp,fp,# */
     if(x)
-      *(unsigned long *)(cur_text_section->data + func_sub_sp_offset) = x;
+      *(uint32_t *)(cur_text_section->data + func_sub_sp_offset) = x;
     else {
-      unsigned long addr;
+      int addr;
       addr=ind;
       o(0xE59FC004); /* ldr ip,[pc+4] */
       o(0xE04BD00C); /* sub sp,fp,ip  */
       o(0xE1A0F00E); /* mov pc,lr */
       o(diff);
-      *(unsigned long *)(cur_text_section->data + func_sub_sp_offset) = 0xE1000000|encbranch(func_sub_sp_offset,addr,1);
+      *(uint32_t *)(cur_text_section->data + func_sub_sp_offset) = 0xE1000000|encbranch(func_sub_sp_offset,addr,1);
     }
   }
 }
@@ -1032,7 +1301,7 @@ void gjmp_addr(int a)
 int gtst(int inv, int t)
 {
   int v, r;
-  unsigned long op;
+  uint32_t op;
   v = vtop->r & VT_VALMASK;
   r=ind;
   if (v == VT_CMP) {
@@ -1045,14 +1314,14 @@ int gtst(int inv, int t)
       if(!vtop->c.i)
 	vtop->c.i=t;
       else {
-	unsigned long *x;
+	uint32_t *x;
 	int p,lp;
 	if(t) {
           p = vtop->c.i;
           do {
 	    p = decbranch(lp=p);
           } while(p);
-	  x = (unsigned long *)(cur_text_section->data + lp);
+	  x = (uint32_t *)(cur_text_section->data + lp);
 	  *x &= 0xff000000;
 	  *x |= encbranch(lp,t,1);
 	}
@@ -1094,7 +1363,7 @@ int gtst(int inv, int t)
 void gen_opi(int op)
 {
   int c, func = 0;
-  unsigned long opc = 0,r,fr;
+  uint32_t opc = 0, r, fr;
   unsigned short retreg = REG_IRET;
 
   c=0;
@@ -1210,7 +1479,7 @@ void gen_opi(int op)
       vswap();
       opc=0xE0000000|(opc<<20)|(c<<16);
       if((vtop->r & (VT_VALMASK | VT_LVAL | VT_SYM)) == VT_CONST) {
-	unsigned long x;
+	uint32_t x;
 	x=stuff_const(opc|0x2000000,vtop->c.i);
 	if(x) {
 	  r=intr(vtop[-1].r=get_reg_ex(RC_INT,regmask(vtop[-1].r)));
@@ -1256,7 +1525,7 @@ done:
       vtop->r = retreg;
       break;
     default:
-      error("gen_opi %i unimplemented!",op);
+      tcc_error("gen_opi %i unimplemented!",op);
   }
 }
 
@@ -1276,7 +1545,7 @@ static int is_zero(int i)
  *    two operands are guaranted to have the same floating point type */
 void gen_opf(int op)
 {
-  unsigned long x;
+  uint32_t x;
   int fneg=0,r;
   x=0xEE000A00|T2CPR(vtop->type.t);
   switch(op) {
@@ -1309,8 +1578,8 @@ void gen_opf(int op)
       x|=0x800000;
       break;
     default:
-      if(op < TOK_ULT && op > TOK_GT) {
-        error("unknown fp op %x!",op);
+      if(op < TOK_ULT || op > TOK_GT) {
+        tcc_error("unknown fp op %x!",op);
         return;
       }
       if(is_zero(-1)) {
@@ -1364,10 +1633,10 @@ void gen_opf(int op)
 }
 
 #else
-static int is_fconst()
+static uint32_t is_fconst()
 {
   long double f;
-  int r;
+  uint32_t r;
   if((vtop->r & (VT_VALMASK | VT_LVAL | VT_SYM)) != VT_CONST)
     return 0;
   if (vtop->type.t == VT_FLOAT)
@@ -1406,8 +1675,7 @@ static int is_fconst()
    two operands are guaranted to have the same floating point type */
 void gen_opf(int op)
 {
-  unsigned long x;
-  int r,r2,c1,c2;
+  uint32_t x, r, r2, c1, c2;
   //fputs("gen_opf\n",stderr);
   vswap();
   c1 = is_fconst();
@@ -1506,7 +1774,7 @@ void gen_opf(int op)
 	  case TOK_UGE:
 	  case TOK_ULE:
 	  case TOK_UGT:
-            error("unsigned comparision on floats?");
+            tcc_error("unsigned comparision on floats?");
 	    break;
 	  case TOK_LT:
             op=TOK_Nset;
@@ -1550,7 +1818,7 @@ void gen_opf(int op)
 	vtop[-1].r = VT_CMP;
 	vtop[-1].c.i = op;
       } else {
-        error("unknown fp op %x!",op);
+        tcc_error("unknown fp op %x!",op);
 	return;
       }
   }
@@ -1570,19 +1838,20 @@ void gen_opf(int op)
 
 /* convert integers to fp 't' type. Must handle 'int', 'unsigned int'
    and 'long long' cases. */
-void gen_cvt_itof1(int t)
+ST_FUNC void gen_cvt_itof1(int t)
 {
-  int r,r2,bt;
+  uint32_t r, r2;
+  int bt;
   bt=vtop->type.t & VT_BTYPE;
   if(bt == VT_INT || bt == VT_SHORT || bt == VT_BYTE) {
 #ifndef TCC_ARM_VFP
-    unsigned int dsize=0;
+    uint32_t dsize = 0;
 #endif
     r=intr(gv(RC_INT));
 #ifdef TCC_ARM_VFP
     r2=vfpr(vtop->r=get_reg(RC_FLOAT));
     o(0xEE000A10|(r<<12)|(r2<<16)); /* fmsr */
-    r2<<=12;
+    r2|=r2<<12;
     if(!(vtop->type.t & VT_UNSIGNED))
       r2|=0x80;                /* fuitoX -> fsituX */
     o(0xEEB80A40|r2|T2CPR(t)); /* fYitoX*/
@@ -1592,7 +1861,7 @@ void gen_cvt_itof1(int t)
       dsize=0x80;    /* flts -> fltd */
     o(0xEE000110|dsize|(r2<<16)|(r<<12)); /* flts */
     if((vtop->type.t & (VT_UNSIGNED|VT_BTYPE)) == (VT_UNSIGNED|VT_INT)) {
-      unsigned int off=0;
+      uint32_t off = 0;
       o(0xE3500000|(r<<12));        /* cmp */
       r=fpr(get_reg(RC_FLOAT));
       if(last_itod_magic) {
@@ -1646,13 +1915,14 @@ void gen_cvt_itof1(int t)
       return;
     }
   }
-  error("unimplemented gen_cvt_itof %x!",vtop->type.t);
+  tcc_error("unimplemented gen_cvt_itof %x!",vtop->type.t);
 }
 
 /* convert fp to int 't' type */
 void gen_cvt_ftoi(int t)
 {
-  int r,r2,u,func=0;
+  uint32_t r, r2;
+  int u, func = 0;
   u=t&VT_UNSIGNED;
   t&=VT_BTYPE;
   r2=vtop->type.t & VT_BTYPE;
@@ -1660,7 +1930,7 @@ void gen_cvt_ftoi(int t)
 #ifdef TCC_ARM_VFP
     r=vfpr(gv(RC_FLOAT));
     u=u?0:0x10000;
-    o(0xEEBC0A40|(r<<12)|r|T2CPR(r2)); /* ftoXiY */
+    o(0xEEBC0AC0|(r<<12)|r|T2CPR(r2)|u); /* ftoXizY */
     r2=intr(vtop->r=get_reg(RC_INT));
     o(0xEE100A10|(r<<16)|(r2<<12));
     return;
@@ -1705,7 +1975,7 @@ void gen_cvt_ftoi(int t)
     vtop->r = REG_IRET;
     return;
   }
-  error("unimplemented gen_cvt_ftoi!");
+  tcc_error("unimplemented gen_cvt_ftoi!");
 }
 
 /* convert from one floating point type to another */
@@ -1713,7 +1983,7 @@ void gen_cvt_ftof(int t)
 {
 #ifdef TCC_ARM_VFP
   if(((vtop->type.t & VT_BTYPE) == VT_FLOAT) != ((t & VT_BTYPE) == VT_FLOAT)) {
-    int r=vfpr(gv(RC_FLOAT));
+    uint32_t r = vfpr(gv(RC_FLOAT));
     o(0xEEB70AC0|(r<<12)|r|T2CPR(vtop->type.t));
   }
 #else
@@ -1731,4 +2001,5 @@ void ggoto(void)
 
 /* end of ARM code generator */
 /*************************************************************/
-
+#endif
+/*************************************************************/
